from __future__ import annotations

import shutil
import uuid
from pathlib import Path

from cliproom.cli import _print_status
from cliproom.cli_state import LocalCLIState


def _mk_state_dir() -> Path:
    path = Path(".test_state") / str(uuid.uuid4())
    path.mkdir(parents=True, exist_ok=True)
    return path


def test_status_no_state(capsys, monkeypatch) -> None:
    state_dir = _mk_state_dir()
    monkeypatch.setenv("CLIPROOM_STATE_DIR", str(state_dir))
    try:
        _print_status()
        out = capsys.readouterr().out
        assert "No local runtime state found." in out
    finally:
        shutil.rmtree(state_dir, ignore_errors=True)


def test_status_client_and_server(capsys, monkeypatch) -> None:
    state_dir = _mk_state_dir()
    monkeypatch.setenv("CLIPROOM_STATE_DIR", str(state_dir))
    try:
        state = LocalCLIState()
        state.write_state(
            {
                "active": True,
                "mode": "join",
                "status": "connected",
                "session": "devteam",
                "server_host": "127.0.0.1",
                "server_port": 4545,
                "client_name": "laptop",
                "pid": 999999,
                "updated_at": "now",
                "last_error": None,
            }
        )
        state.write_server_state(
            {
                "active": True,
                "status": "running",
                "fixed_session": "devteam",
                "host": "0.0.0.0",
                "port": 4545,
                "pid": 999998,
                "updated_at": "now",
                "last_error": None,
            }
        )

        _print_status()
        out = capsys.readouterr().out
        assert "[client]" in out
        assert "[server]" in out
        assert "session: devteam" in out
        assert "endpoint: 127.0.0.1:4545" in out
        assert "endpoint: 0.0.0.0:4545" in out
    finally:
        shutil.rmtree(state_dir, ignore_errors=True)



"""Session and client state management for ClipRoom server."""

from __future__ import annotations

import asyncio
from dataclasses import dataclass, field

from cliproom.protocol import ProtocolValidationError


@dataclass(slots=True, eq=False)
class ClientConnection:
    """A connected client bound to a stream writer."""

    client_name: str
    writer: asyncio.StreamWriter
    lock: asyncio.Lock = field(default_factory=asyncio.Lock)

    async def send(self, payload: bytes) -> None:
        """Write a pre-encoded CRP message to this client."""
        async with self.lock:
            self.writer.write(payload)
            await self.writer.drain()


@dataclass(slots=True)
class SessionState:
    """A logical room identified by session name."""

    name: str
    auth_hash: str
    clients: set[ClientConnection] = field(default_factory=set)

    def validate_auth(self, auth_hash: str) -> None:
        if self.auth_hash != auth_hash:
            raise ProtocolValidationError("invalid_auth")


class SessionManager:
    """Owns all active sessions."""

    def __init__(
        self,
        *,
        fixed_session: str | None = None,
        fixed_auth_hash: str | None = None,
    ) -> None:
        self._sessions: dict[str, SessionState] = {}
        self._lock = asyncio.Lock()
        self._fixed_session = fixed_session
        self._fixed_auth_hash = fixed_auth_hash or ""

    async def join(
        self, session_name: str, auth_hash: str, connection: ClientConnection
    ) -> SessionState:
        """Join or create a session and register the client."""
        async with self._lock:
            if self._fixed_session and session_name != self._fixed_session:
                raise ProtocolValidationError("invalid_session")

            if self._fixed_session:
                session = self._sessions.get(self._fixed_session)
                if session is None:
                    session = SessionState(
                        name=self._fixed_session,
                        auth_hash=self._fixed_auth_hash,
                    )
                    self._sessions[self._fixed_session] = session
            else:
                session = self._sessions.get(session_name)
                if session is None:
                    session = SessionState(name=session_name, auth_hash=auth_hash)
                    self._sessions[session_name] = session

            session.validate_auth(auth_hash)
            session.clients.add(connection)
            return session

    async def leave(self, session_name: str, connection: ClientConnection) -> None:
        """Remove a client from a session and cleanup empty sessions."""
        async with self._lock:
            session = self._sessions.get(session_name)
            if session is None:
                return
            session.clients.discard(connection)
            if not session.clients:
                self._sessions.pop(session_name, None)

    async def list_clients(self, session_name: str) -> tuple[ClientConnection, ...]:
        """Return a snapshot of session clients."""
        async with self._lock:
            session = self._sessions.get(session_name)
            if session is None:
                return ()
            return tuple(session.clients)

    async def list_discoverable_sessions(self) -> tuple[tuple[str, bool], ...]:
        """Return discoverable sessions as (name, requires_passphrase)."""
        async with self._lock:
            advertised: list[tuple[str, bool]] = []

            if self._fixed_session:
                advertised.append((self._fixed_session, bool(self._fixed_auth_hash)))

            for name, session in self._sessions.items():
                if self._fixed_session and name == self._fixed_session:
                    continue
                advertised.append((name, bool(session.auth_hash)))

            return tuple(advertised)



"""Serialization and parsing helpers for CRP messages."""

from __future__ import annotations

import hashlib
import json
from dataclasses import asdict

from .constants import CRP_VERSION
from .errors import ProtocolValidationError
from .messages import (
    ClipboardFormat,
    ClipMessage,
    DiscoverMessage,
    DiscoverReplyMessage,
    ErrorMessage,
    HeartbeatMessage,
    JoinAckMessage,
    JoinMessage,
    LeaveMessage,
)
from .types import CRPMessage, JSONObject


def hash_passphrase(passphrase: str) -> str:
    """Return SHA256(passphrase) in hex format."""
    if not isinstance(passphrase, str):
        raise ProtocolValidationError("'passphrase' must be string")
    return hashlib.sha256(passphrase.encode("utf-8")).hexdigest()


def _validate_protocol(payload: JSONObject) -> None:
    protocol = payload.get("protocol")
    if protocol is None:
        return
    if protocol != CRP_VERSION:
        raise ProtocolValidationError(
            f"Unsupported protocol '{protocol}', expected '{CRP_VERSION}'"
        )


def parse_message(payload: JSONObject) -> CRPMessage:
    """Convert a dict payload into a typed CRP message."""
    if not isinstance(payload, dict):
        raise ProtocolValidationError("Message payload must be a JSON object")

    _validate_protocol(payload)

    message_type = payload.get("type")
    if message_type == "join":
        return JoinMessage(
            session=payload["session"],
            auth=payload["auth"],
            client_name=payload["client_name"],
        )
    if message_type == "join_ack":
        return JoinAckMessage(session=payload["session"])
    if message_type == "clip":
        raw_formats = payload.get("formats", [])
        if not isinstance(raw_formats, list):
            raise ProtocolValidationError("'formats' must be a list")
        return ClipMessage(
            session=payload["session"],
            message_id=payload["message_id"],
            formats=[
                ClipboardFormat(
                    mime=item["mime"],
                    encoding=item["encoding"],
                    data=item["data"],
                )
                for item in raw_formats
            ],
        )
    if message_type == "leave":
        return LeaveMessage()
    if message_type == "heartbeat":
        return HeartbeatMessage()
    if message_type == "error":
        return ErrorMessage(message=payload["message"])
    if message_type == "discover":
        return DiscoverMessage()
    if message_type == "discover_reply":
        return DiscoverReplyMessage(
            session=payload["session"],
            requires_passphrase=payload["requires_passphrase"],
            host=payload["host"],
            port=payload["port"],
        )

    raise ProtocolValidationError(f"Unsupported message type '{message_type}'")


def message_to_dict(message: CRPMessage) -> JSONObject:
    """Convert a typed CRP message into a JSON-ready dictionary."""
    return asdict(message)


def encode_message(message: CRPMessage) -> bytes:
    """Encode message as UTF-8 newline-delimited JSON."""
    return (json.dumps(message_to_dict(message), separators=(",", ":")) + "\n").encode(
        "utf-8"
    )


def decode_message_line(line: str | bytes) -> CRPMessage:
    """Decode one newline-delimited JSON message."""
    if isinstance(line, bytes):
        line = line.decode("utf-8")
    line = line.strip()
    if not line:
        raise ProtocolValidationError("Empty message line")
    try:
        payload = json.loads(line)
    except json.JSONDecodeError as exc:
        raise ProtocolValidationError(f"Invalid JSON message: {exc}") from exc
    return parse_message(payload)



"""Clipboard adapters."""

from .adapters import (
    ClipboardAdapter,
    ClipboardUnavailableError,
    SUPPORTED_MIME_TYPES,
    create_default_adapter,
)

__all__ = [
    "ClipboardAdapter",
    "ClipboardUnavailableError",
    "SUPPORTED_MIME_TYPES",
    "create_default_adapter",
]



from __future__ import annotations

from cliproom.protocol import (
    ClipboardFormat,
    ClipMessage,
    JoinMessage,
    decode_message_line,
    encode_message,
    hash_passphrase,
)


def test_join_roundtrip() -> None:
    join = JoinMessage(
        session="devteam",
        auth=hash_passphrase("secret"),
        client_name="laptop",
    )
    raw = encode_message(join)
    parsed = decode_message_line(raw)
    assert parsed.type == "join"
    assert parsed.session == "devteam"
    assert parsed.client_name == "laptop"


def test_clip_roundtrip() -> None:
    clip = ClipMessage(
        session="devteam",
        message_id="m1",
        formats=[
            ClipboardFormat(
                mime="text/plain",
                encoding="utf-8",
                data="SGVsbG8=",
            )
        ],
    )
    raw = encode_message(clip)
    parsed = decode_message_line(raw)
    assert parsed.type == "clip"
    assert parsed.message_id == "m1"
    assert parsed.formats[0].mime == "text/plain"



from __future__ import annotations

import asyncio

import pytest

from cliproom.protocol import ProtocolValidationError
from cliproom.server.session_manager import ClientConnection, SessionManager


class _DummyWriter:
    def write(self, payload: bytes) -> None:
        _ = payload

    async def drain(self) -> None:
        return


def _conn(name: str) -> ClientConnection:
    return ClientConnection(client_name=name, writer=_DummyWriter())  # type: ignore[arg-type]


def test_join_and_discoverable_sessions() -> None:
    async def scenario() -> None:
        mgr = SessionManager()
        c1 = _conn("a")
        c2 = _conn("b")
        s = await mgr.join("devteam", "h1", c1)
        assert s.name == "devteam"
        await mgr.join("support", "h2", c2)
        sessions = await mgr.list_discoverable_sessions()
        assert ("devteam", True) in sessions
        assert ("support", True) in sessions

    asyncio.run(scenario())


def test_invalid_auth_rejected() -> None:
    async def scenario() -> None:
        mgr = SessionManager()
        await mgr.join("devteam", "h1", _conn("a"))
        with pytest.raises(ProtocolValidationError):
            await mgr.join("devteam", "h2", _conn("b"))

    asyncio.run(scenario())



"""Backward-compatible re-exports for protocol message models."""

from .errors import ProtocolValidationError
from .messages import (
    ClipboardFormat,
    ClipMessage,
    DiscoverMessage,
    DiscoverReplyMessage,
    ErrorMessage,
    HeartbeatMessage,
    JoinAckMessage,
    JoinMessage,
    LeaveMessage,
)
from .types import CRPMessage

__all__ = [
    "ProtocolValidationError",
    "ClipboardFormat",
    "JoinMessage",
    "JoinAckMessage",
    "ClipMessage",
    "LeaveMessage",
    "HeartbeatMessage",
    "ErrorMessage",
    "DiscoverMessage",
    "DiscoverReplyMessage",
    "CRPMessage",
]



"""Typed CRP v0.1 message contracts."""

from __future__ import annotations

from dataclasses import dataclass, field
from typing import Literal

from .constants import CRP_VERSION
from .errors import ProtocolValidationError


def _require_non_empty(value: str, field_name: str) -> None:
    if not isinstance(value, str) or not value.strip():
        raise ProtocolValidationError(f"'{field_name}' must be a non-empty string")


@dataclass(slots=True)
class ClipboardFormat:
    mime: str
    encoding: str
    data: str

    def __post_init__(self) -> None:
        _require_non_empty(self.mime, "mime")
        _require_non_empty(self.encoding, "encoding")
        _require_non_empty(self.data, "data")


@dataclass(slots=True)
class JoinMessage:
    session: str
    auth: str
    client_name: str
    protocol: str = field(default=CRP_VERSION, init=False)
    type: Literal["join"] = field(default="join", init=False)

    def __post_init__(self) -> None:
        _require_non_empty(self.session, "session")
        _require_non_empty(self.auth, "auth")
        _require_non_empty(self.client_name, "client_name")


@dataclass(slots=True)
class JoinAckMessage:
    session: str
    protocol: str = field(default=CRP_VERSION, init=False)
    type: Literal["join_ack"] = field(default="join_ack", init=False)

    def __post_init__(self) -> None:
        _require_non_empty(self.session, "session")


@dataclass(slots=True)
class ClipMessage:
    session: str
    message_id: str
    formats: list[ClipboardFormat]
    protocol: str = field(default=CRP_VERSION, init=False)
    type: Literal["clip"] = field(default="clip", init=False)

    def __post_init__(self) -> None:
        _require_non_empty(self.session, "session")
        _require_non_empty(self.message_id, "message_id")
        if not isinstance(self.formats, list) or not self.formats:
            raise ProtocolValidationError("'formats' must be a non-empty list")


@dataclass(slots=True)
class LeaveMessage:
    protocol: str = field(default=CRP_VERSION, init=False)
    type: Literal["leave"] = field(default="leave", init=False)


@dataclass(slots=True)
class HeartbeatMessage:
    protocol: str = field(default=CRP_VERSION, init=False)
    type: Literal["heartbeat"] = field(default="heartbeat", init=False)


@dataclass(slots=True)
class ErrorMessage:
    message: str
    protocol: str = field(default=CRP_VERSION, init=False)
    type: Literal["error"] = field(default="error", init=False)

    def __post_init__(self) -> None:
        _require_non_empty(self.message, "message")


@dataclass(slots=True)
class DiscoverMessage:
    protocol: str = field(default=CRP_VERSION, init=False)
    type: Literal["discover"] = field(default="discover", init=False)


@dataclass(slots=True)
class DiscoverReplyMessage:
    session: str
    requires_passphrase: bool
    host: str
    port: int
    protocol: str = field(default=CRP_VERSION, init=False)
    type: Literal["discover_reply"] = field(default="discover_reply", init=False)

    def __post_init__(self) -> None:
        _require_non_empty(self.session, "session")
        _require_non_empty(self.host, "host")
        if not isinstance(self.requires_passphrase, bool):
            raise ProtocolValidationError("'requires_passphrase' must be boolean")
        if not isinstance(self.port, int) or self.port <= 0:
            raise ProtocolValidationError("'port' must be a positive integer")




"""Async TCP server runtime for CRP."""

from __future__ import annotations

import asyncio
import logging
import socket
from contextlib import suppress

from cliproom.protocol import (
    ClipMessage,
    ErrorMessage,
    HeartbeatMessage,
    DiscoverMessage,
    DiscoverReplyMessage,
    JoinAckMessage,
    JoinMessage,
    LeaveMessage,
    ProtocolValidationError,
    decode_message_line,
    encode_message,
    DEFAULT_DISCOVERY_PORT,
)

from .session_manager import ClientConnection, SessionManager


LOG = logging.getLogger("cliproom.server")


class ClipRoomServer:
    """Stateful CRP server with session routing."""

    def __init__(
        self,
        host: str,
        port: int,
        *,
        fixed_session: str | None = None,
        fixed_auth_hash: str | None = None,
        discovery_port: int = DEFAULT_DISCOVERY_PORT,
        enable_discovery: bool = True,
    ) -> None:
        self.host = host
        self.port = port
        self.discovery_port = discovery_port
        self.enable_discovery = enable_discovery
        self._session_manager = SessionManager(
            fixed_session=fixed_session,
            fixed_auth_hash=fixed_auth_hash,
        )
        self._server: asyncio.AbstractServer | None = None
        self._discovery_transport: asyncio.DatagramTransport | None = None

    async def start(self) -> None:
        self._server = await asyncio.start_server(self._handle_client, self.host, self.port)
        sockets = self._server.sockets or []
        addresses = ", ".join(str(sock.getsockname()) for sock in sockets)
        LOG.info("Server listening on %s", addresses)
        if self.enable_discovery:
            await self._start_discovery()

    async def serve_forever(self) -> None:
        if self._server is None:
            await self.start()
        assert self._server is not None
        async with self._server:
            await self._server.serve_forever()

    async def stop(self) -> None:
        if self._server is None:
            return
        self._server.close()
        await self._server.wait_closed()
        self._server = None
        if self._discovery_transport is not None:
            self._discovery_transport.close()
            self._discovery_transport = None

    async def _handle_client(
        self,
        reader: asyncio.StreamReader,
        writer: asyncio.StreamWriter,
    ) -> None:
        peer = writer.get_extra_info("peername")
        connection: ClientConnection | None = None
        session_name: str | None = None
        LOG.info("Client connected: %s", peer)
        try:
            while True:
                raw_line = await reader.readline()
                if not raw_line:
                    break

                message = decode_message_line(raw_line)
                if connection is None:
                    if not isinstance(message, JoinMessage):
                        raise ProtocolValidationError("join_required")

                    connection = ClientConnection(
                        client_name=message.client_name,
                        writer=writer,
                    )
                    session = await self._session_manager.join(
                        session_name=message.session,
                        auth_hash=message.auth,
                        connection=connection,
                    )
                    session_name = session.name
                    await connection.send(encode_message(JoinAckMessage(session=session.name)))
                    LOG.info("Client %s joined session %s", message.client_name, session_name)
                    continue

                if isinstance(message, ClipMessage):
                    if message.session != session_name:
                        raise ProtocolValidationError("invalid_session")
                    await self._broadcast(session_name, connection, encode_message(message))
                    continue

                if isinstance(message, (HeartbeatMessage,)):
                    continue

                if isinstance(message, LeaveMessage):
                    break

                raise ProtocolValidationError(f"unsupported_after_join:{message.type}")
        except ProtocolValidationError as exc:
            LOG.warning("Protocol error from %s: %s", peer, exc)
            await self._safe_send_error(writer, str(exc))
        except Exception:
            LOG.exception("Unhandled error serving client %s", peer)
            await self._safe_send_error(writer, "internal_error")
        finally:
            if connection is not None and session_name is not None:
                await self._session_manager.leave(session_name, connection)
            writer.close()
            with suppress(Exception):
                await writer.wait_closed()
            LOG.info("Client disconnected: %s", peer)

    async def _broadcast(
        self,
        session_name: str,
        origin: ClientConnection,
        payload: bytes,
    ) -> None:
        targets = [
            client
            for client in await self._session_manager.list_clients(session_name)
            if client is not origin
        ]
        if not targets:
            return
        await asyncio.gather(*(client.send(payload) for client in targets), return_exceptions=True)

    async def _safe_send_error(self, writer: asyncio.StreamWriter, message: str) -> None:
        try:
            writer.write(encode_message(ErrorMessage(message=message)))
            await writer.drain()
        except Exception:
            pass

    async def _start_discovery(self) -> None:
        loop = asyncio.get_running_loop()
        transport, _ = await loop.create_datagram_endpoint(
            lambda: _DiscoveryServerProtocol(self),
            local_addr=("0.0.0.0", self.discovery_port),
            allow_broadcast=True,
        )
        self._discovery_transport = transport
        LOG.info("Discovery listening on UDP %s", self.discovery_port)

    async def _build_discovery_replies(
        self, requester_addr: tuple[str, int]
    ) -> list[bytes]:
        host = self._advertise_host_for(requester_addr[0])
        replies: list[bytes] = []
        for session_name, requires_passphrase in (
            await self._session_manager.list_discoverable_sessions()
        ):
            reply = DiscoverReplyMessage(
                session=session_name,
                requires_passphrase=requires_passphrase,
                host=host,
                port=self.port,
            )
            replies.append(encode_message(reply))
        return replies

    def _advertise_host_for(self, requester_host: str) -> str:
        if self.host not in {"0.0.0.0", "::"}:
            return self.host
        sock = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
        try:
            sock.connect((requester_host, 9))
            local_ip = sock.getsockname()[0]
            return local_ip
        except OSError:
            return "127.0.0.1"
        finally:
            sock.close()


class _DiscoveryServerProtocol(asyncio.DatagramProtocol):
    def __init__(self, server: ClipRoomServer) -> None:
        self.server = server
        self.transport: asyncio.DatagramTransport | None = None

    def connection_made(self, transport: asyncio.BaseTransport) -> None:
        self.transport = transport  # type: ignore[assignment]

    def datagram_received(self, data: bytes, addr: tuple[str, int]) -> None:
        if self.transport is None:
            return
        try:
            message = decode_message_line(data)
            if not isinstance(message, DiscoverMessage):
                return
        except Exception:
            return
        asyncio.create_task(self._reply_to_discover(addr))

    async def _reply_to_discover(self, addr: tuple[str, int]) -> None:
        if self.transport is None:
            return
        for payload in await self.server._build_discovery_replies(addr):
            self.transport.sendto(payload, addr)


async def run_server(
    host: str,
    port: int,
    *,
    fixed_session: str | None = None,
    fixed_auth_hash: str | None = None,
    discovery_port: int = DEFAULT_DISCOVERY_PORT,
    enable_discovery: bool = True,
) -> None:
    """Create and run a server process until interrupted."""
    server = ClipRoomServer(
        host=host,
        port=port,
        fixed_session=fixed_session,
        fixed_auth_hash=fixed_auth_hash,
        discovery_port=discovery_port,
        enable_discovery=enable_discovery,
    )
    await server.serve_forever()



"""Clipboard adapters for supported platforms."""

from __future__ import annotations

import base64
import shutil
import subprocess
import sys
from abc import ABC, abstractmethod

from cliproom.protocol import ClipboardFormat


SUPPORTED_MIME_TYPES = (
    "text/plain",
    "text/html",
    "image/png",
    "image/jpeg",
    "application/x-file-list",
)


class ClipboardUnavailableError(RuntimeError):
    """Raised when no compatible clipboard backend is available."""


class ClipboardAdapter(ABC):
    """Platform clipboard abstraction with multi-format support."""

    @abstractmethod
    def get_formats(self) -> list[ClipboardFormat]:
        """Read available clipboard formats in protocol representation."""

    @abstractmethod
    def set_formats(self, formats: list[ClipboardFormat]) -> None:
        """Write clipboard formats using best-effort MIME preference."""

    def get_text(self) -> str:
        for item in self.get_formats():
            if item.mime == "text/plain":
                return base64.b64decode(item.data).decode("utf-8")
        return ""

    def set_text(self, value: str) -> None:
        encoded = base64.b64encode(value.encode("utf-8")).decode("ascii")
        self.set_formats(
            [ClipboardFormat(mime="text/plain", encoding="utf-8", data=encoded)]
        )


def _to_clipboard_format(mime: str, payload: bytes, *, text: bool) -> ClipboardFormat:
    if text:
        encoded = base64.b64encode(payload).decode("ascii")
        return ClipboardFormat(mime=mime, encoding="utf-8", data=encoded)
    encoded = base64.b64encode(payload).decode("ascii")
    return ClipboardFormat(mime=mime, encoding="base64", data=encoded)


def _decode_clipboard_data(item: ClipboardFormat) -> bytes:
    return base64.b64decode(item.data)


class WindowsClipboardAdapter(ClipboardAdapter):
    def get_formats(self) -> list[ClipboardFormat]:
        formats: list[ClipboardFormat] = []

        plain = self._run_ps("Get-Clipboard -Raw")
        if plain is not None and plain != "":
            formats.append(_to_clipboard_format("text/plain", plain.encode("utf-8"), text=True))

        html = self._run_ps("Get-Clipboard -Format Html -Raw")
        if html:
            formats.append(_to_clipboard_format("text/html", html.encode("utf-8"), text=True))

        file_list = self._run_ps("(Get-Clipboard -Format FileDropList) -join \"`n\"")
        if file_list:
            formats.append(
                _to_clipboard_format(
                    "application/x-file-list",
                    file_list.encode("utf-8"),
                    text=True,
                )
            )

        png_b64 = self._run_ps(
            "$img=Get-Clipboard -Format Image;"
            "if($img){$ms=New-Object System.IO.MemoryStream;"
            "$img.Save($ms,[System.Drawing.Imaging.ImageFormat]::Png);"
            "[Convert]::ToBase64String($ms.ToArray())}"
        )
        if png_b64:
            with_bytes = base64.b64decode(png_b64)
            formats.append(_to_clipboard_format("image/png", with_bytes, text=False))

        return formats

    def set_formats(self, formats: list[ClipboardFormat]) -> None:
        by_mime = {f.mime: f for f in formats}

        if "image/png" in by_mime:
            payload = base64.b64encode(_decode_clipboard_data(by_mime["image/png"])).decode("ascii")
            script = (
                "$b=[Convert]::FromBase64String([Console]::In.ReadToEnd());"
                "$ms=New-Object IO.MemoryStream(,$b);"
                "$img=[System.Drawing.Image]::FromStream($ms);"
                "Set-Clipboard -Value $img"
            )
            self._run_ps_stdin(script, payload)
            return

        if "image/jpeg" in by_mime:
            payload = base64.b64encode(_decode_clipboard_data(by_mime["image/jpeg"])).decode("ascii")
            script = (
                "$b=[Convert]::FromBase64String([Console]::In.ReadToEnd());"
                "$ms=New-Object IO.MemoryStream(,$b);"
                "$img=[System.Drawing.Image]::FromStream($ms);"
                "Set-Clipboard -Value $img"
            )
            self._run_ps_stdin(script, payload)
            return

        if "text/html" in by_mime:
            text = _decode_clipboard_data(by_mime["text/html"]).decode("utf-8")
            self._run_ps_stdin("Set-Clipboard -AsHtml ([Console]::In.ReadToEnd())", text)
            return

        if "application/x-file-list" in by_mime:
            text = _decode_clipboard_data(by_mime["application/x-file-list"]).decode("utf-8")
            self._run_ps_stdin("Set-Clipboard -Value ([Console]::In.ReadToEnd())", text)
            return

        if "text/plain" in by_mime:
            text = _decode_clipboard_data(by_mime["text/plain"]).decode("utf-8")
            self._run_ps_stdin("Set-Clipboard -Value ([Console]::In.ReadToEnd())", text)
            return

        raise ClipboardUnavailableError("No supported MIME found in payload")

    def _run_ps(self, script: str) -> str | None:
        process = subprocess.run(
            ["powershell", "-NoProfile", "-Command", script],
            text=True,
            capture_output=True,
            check=False,
        )
        if process.returncode != 0:
            return None
        return process.stdout

    def _run_ps_stdin(self, script: str, value: str) -> None:
        process = subprocess.run(
            ["powershell", "-NoProfile", "-Command", script],
            text=True,
            input=value,
            capture_output=True,
            check=False,
        )
        if process.returncode != 0:
            raise ClipboardUnavailableError(process.stderr.strip() or "Set-Clipboard failed")


class WlClipboardAdapter(ClipboardAdapter):
    def get_formats(self) -> list[ClipboardFormat]:
        formats: list[ClipboardFormat] = []
        text = self._read_text("text/plain")
        if text:
            formats.append(_to_clipboard_format("text/plain", text, text=True))

        html = self._read_text("text/html")
        if html:
            formats.append(_to_clipboard_format("text/html", html, text=True))

        png = self._read_bytes("image/png")
        if png:
            formats.append(_to_clipboard_format("image/png", png, text=False))

        jpeg = self._read_bytes("image/jpeg")
        if jpeg:
            formats.append(_to_clipboard_format("image/jpeg", jpeg, text=False))

        file_list = self._read_text("text/uri-list")
        if file_list:
            formats.append(
                _to_clipboard_format("application/x-file-list", file_list, text=True)
            )
        return formats

    def set_formats(self, formats: list[ClipboardFormat]) -> None:
        by_mime = {f.mime: f for f in formats}
        for mime in ("image/png", "image/jpeg", "text/html", "application/x-file-list", "text/plain"):
            item = by_mime.get(mime)
            if item is None:
                continue
            data = _decode_clipboard_data(item)
            target = "text/uri-list" if mime == "application/x-file-list" else mime
            process = subprocess.run(
                ["wl-copy", "--type", target],
                input=data,
                capture_output=True,
                check=False,
            )
            if process.returncode != 0:
                raise ClipboardUnavailableError(
                    process.stderr.decode("utf-8", errors="ignore").strip()
                    or "wl-copy failed"
                )
            return
        raise ClipboardUnavailableError("No supported MIME found in payload")

    def _read_text(self, mime: str) -> bytes | None:
        process = subprocess.run(
            ["wl-paste", "--no-newline", "--type", mime],
            capture_output=True,
            check=False,
        )
        if process.returncode != 0 or not process.stdout:
            return None
        return process.stdout

    def _read_bytes(self, mime: str) -> bytes | None:
        process = subprocess.run(
            ["wl-paste", "--type", mime],
            capture_output=True,
            check=False,
        )
        if process.returncode != 0 or not process.stdout:
            return None
        return process.stdout


class XclipClipboardAdapter(ClipboardAdapter):
    def get_formats(self) -> list[ClipboardFormat]:
        targets = self._targets()
        formats: list[ClipboardFormat] = []

        if "UTF8_STRING" in targets or "text/plain" in targets:
            text = self._read("text/plain")
            if text:
                formats.append(_to_clipboard_format("text/plain", text, text=True))

        if "text/html" in targets:
            html = self._read("text/html")
            if html:
                formats.append(_to_clipboard_format("text/html", html, text=True))

        if "image/png" in targets:
            png = self._read("image/png")
            if png:
                formats.append(_to_clipboard_format("image/png", png, text=False))

        if "image/jpeg" in targets:
            jpeg = self._read("image/jpeg")
            if jpeg:
                formats.append(_to_clipboard_format("image/jpeg", jpeg, text=False))

        if "text/uri-list" in targets:
            file_list = self._read("text/uri-list")
            if file_list:
                formats.append(
                    _to_clipboard_format("application/x-file-list", file_list, text=True)
                )
        return formats

    def set_formats(self, formats: list[ClipboardFormat]) -> None:
        by_mime = {f.mime: f for f in formats}
        mapping = (
            ("image/png", "image/png"),
            ("image/jpeg", "image/jpeg"),
            ("text/html", "text/html"),
            ("application/x-file-list", "text/uri-list"),
            ("text/plain", "text/plain"),
        )
        for source_mime, target in mapping:
            item = by_mime.get(source_mime)
            if item is None:
                continue
            data = _decode_clipboard_data(item)
            process = subprocess.run(
                ["xclip", "-selection", "clipboard", "-t", target, "-in"],
                input=data,
                capture_output=True,
                check=False,
            )
            if process.returncode != 0:
                raise ClipboardUnavailableError(
                    process.stderr.decode("utf-8", errors="ignore").strip()
                    or "xclip write failed"
                )
            return
        raise ClipboardUnavailableError("No supported MIME found in payload")

    def _targets(self) -> set[str]:
        process = subprocess.run(
            ["xclip", "-selection", "clipboard", "-t", "TARGETS", "-o"],
            capture_output=True,
            check=False,
        )
        if process.returncode != 0:
            return set()
        return {
            line.strip()
            for line in process.stdout.decode("utf-8", errors="ignore").splitlines()
            if line.strip()
        }

    def _read(self, mime: str) -> bytes | None:
        process = subprocess.run(
            ["xclip", "-selection", "clipboard", "-t", mime, "-o"],
            capture_output=True,
            check=False,
        )
        if process.returncode != 0 or not process.stdout:
            return None
        return process.stdout


def create_default_adapter() -> ClipboardAdapter:
    """Auto-detect best clipboard adapter for the current platform."""
    if sys.platform.startswith("win"):
        return WindowsClipboardAdapter()

    if sys.platform.startswith("linux"):
        if shutil.which("wl-paste") and shutil.which("wl-copy"):
            return WlClipboardAdapter()
        if shutil.which("xclip"):
            return XclipClipboardAdapter()

    raise ClipboardUnavailableError("No supported clipboard backend found")



"""Local CLI state and control signal management."""

from __future__ import annotations

import json
import os
from pathlib import Path
from typing import Any


def _default_state_dir() -> Path:
    env_dir = os.getenv("CLIPROOM_STATE_DIR")
    if env_dir:
        return Path(env_dir)
    return Path.cwd() / ".cliproom"


class LocalCLIState:
    """Stores local runtime status for status/leave commands."""

    def __init__(self, state_dir: Path | None = None) -> None:
        self.state_dir = state_dir or _default_state_dir()
        self.state_file = self.state_dir / "client_state.json"
        self.leave_file = self.state_dir / "leave.request"
        self.server_state_file = self.state_dir / "server_state.json"
        self.server_stop_file = self.state_dir / "server_stop.request"

    def write_state(self, payload: dict[str, Any]) -> None:
        self.state_dir.mkdir(parents=True, exist_ok=True)
        self.state_file.write_text(json.dumps(payload, indent=2), encoding="utf-8")

    def read_state(self) -> dict[str, Any] | None:
        if not self.state_file.exists():
            return None
        try:
            return json.loads(self.state_file.read_text(encoding="utf-8"))
        except Exception:
            return None

    def set_leave_requested(self) -> None:
        self.state_dir.mkdir(parents=True, exist_ok=True)
        self.leave_file.write_text("1", encoding="utf-8")

    def clear_leave_requested(self) -> None:
        if self.leave_file.exists():
            self.leave_file.unlink()

    def is_leave_requested(self) -> bool:
        return self.leave_file.exists()

    def write_server_state(self, payload: dict[str, Any]) -> None:
        self.state_dir.mkdir(parents=True, exist_ok=True)
        self.server_state_file.write_text(json.dumps(payload, indent=2), encoding="utf-8")

    def read_server_state(self) -> dict[str, Any] | None:
        if not self.server_state_file.exists():
            return None
        try:
            return json.loads(self.server_state_file.read_text(encoding="utf-8"))
        except Exception:
            return None

    def set_server_stop_requested(self) -> None:
        self.state_dir.mkdir(parents=True, exist_ok=True)
        self.server_stop_file.write_text("1", encoding="utf-8")

    def clear_server_stop_requested(self) -> None:
        if self.server_stop_file.exists():
            self.server_stop_file.unlink()

    def is_server_stop_requested(self) -> bool:
        return self.server_stop_file.exists()


def is_pid_running(pid: int) -> bool:
    """Return whether pid appears alive."""
    if pid <= 0:
        return False
    try:
        os.kill(pid, 0)
    except OSError:
        return False
    return True



from __future__ import annotations

import asyncio
import socket
import shutil
import uuid
from pathlib import Path

from cliproom.cli import _request_server_stop, _run_server_start
from cliproom.cli_state import LocalCLIState


def _free_port() -> int:
    sock = socket.socket()
    sock.bind(("127.0.0.1", 0))
    port = int(sock.getsockname()[1])
    sock.close()
    return port


def test_server_stop_signal_flow(monkeypatch) -> None:
    state_dir = Path(".test_state") / str(uuid.uuid4())
    state_dir.mkdir(parents=True, exist_ok=True)
    monkeypatch.setenv("CLIPROOM_STATE_DIR", str(state_dir))

    async def scenario() -> None:
        port = _free_port()
        discovery_port = _free_port()
        task = asyncio.create_task(
            _run_server_start(
                host="127.0.0.1",
                port=port,
                fixed_session="devteam",
                fixed_auth_hash="h1",
                discovery_port=discovery_port,
                enable_discovery=False,
            )
        )
        state = LocalCLIState()
        for _ in range(20):
            payload = state.read_server_state() or {}
            if payload.get("status") == "running":
                break
            await asyncio.sleep(0.1)
        payload = state.read_server_state() or {}
        assert payload.get("status") == "running"

        _request_server_stop(wait_seconds=0)
        await asyncio.wait_for(task, timeout=5)

        payload = state.read_server_state() or {}
        assert payload.get("status") == "stopped"
        assert payload.get("active") is False

    try:
        asyncio.run(scenario())
    finally:
        shutil.rmtree(state_dir, ignore_errors=True)



"""Client-side discovery helpers."""

from __future__ import annotations

import asyncio
from dataclasses import dataclass

from cliproom.protocol import (
    DEFAULT_DISCOVERY_PORT,
    DEFAULT_DISCOVERY_TIMEOUT_SECONDS,
    DiscoverMessage,
    DiscoverReplyMessage,
    decode_message_line,
    encode_message,
)


@dataclass(slots=True, frozen=True)
class DiscoveredSession:
    session: str
    host: str
    port: int
    requires_passphrase: bool


class _DiscoveryClientProtocol(asyncio.DatagramProtocol):
    def __init__(self) -> None:
        self.responses: list[DiscoveredSession] = []
        self.transport: asyncio.DatagramTransport | None = None

    def connection_made(self, transport: asyncio.BaseTransport) -> None:
        self.transport = transport  # type: ignore[assignment]

    def datagram_received(self, data: bytes, addr: tuple[str, int]) -> None:
        del addr
        try:
            message = decode_message_line(data)
        except Exception:
            return
        if not isinstance(message, DiscoverReplyMessage):
            return
        self.responses.append(
            DiscoveredSession(
                session=message.session,
                host=message.host,
                port=message.port,
                requires_passphrase=message.requires_passphrase,
            )
        )


async def discover_sessions(
    *,
    broadcast_host: str = "255.255.255.255",
    discovery_port: int = DEFAULT_DISCOVERY_PORT,
    timeout_seconds: float = DEFAULT_DISCOVERY_TIMEOUT_SECONDS,
) -> list[DiscoveredSession]:
    """Broadcast discover and collect replies until timeout."""
    loop = asyncio.get_running_loop()
    protocol = _DiscoveryClientProtocol()
    transport, _ = await loop.create_datagram_endpoint(
        lambda: protocol,
        local_addr=("0.0.0.0", 0),
        allow_broadcast=True,
    )
    try:
        payload = encode_message(DiscoverMessage())
        transport.sendto(payload, (broadcast_host, discovery_port))
        await asyncio.sleep(timeout_seconds)
    finally:
        transport.close()

    unique = {
        (item.session, item.host, item.port, item.requires_passphrase): item
        for item in protocol.responses
    }
    return sorted(unique.values(), key=lambda x: (x.session, x.host, x.port))




"""ClipRoom server package."""

from .runtime import ClipRoomServer, run_server
from .session_manager import ClientConnection, SessionManager, SessionState

__all__ = [
    "ClipRoomServer",
    "run_server",
    "ClientConnection",
    "SessionManager",
    "SessionState",
]



"""CLI entry point for ClipRoom."""

from __future__ import annotations

import argparse
import asyncio
import logging
import os
from datetime import datetime, timezone
from contextlib import suppress

from . import __version__
from .client import ClipRoomClient
from .cli_state import LocalCLIState, is_pid_running
from .discovery import discover_sessions
from .protocol import DEFAULT_DISCOVERY_PORT, DEFAULT_TCP_PORT, hash_passphrase
from .server import ClipRoomServer


def main() -> None:
    """Run ClipRoom CLI."""
    parser = _build_parser()
    args = parser.parse_args()

    if args.command is None:
        print(f"ClipRoom {__version__}")
        parser.print_help()
        return

    logging.basicConfig(
        level=logging.DEBUG if getattr(args, "verbose", False) else logging.INFO,
        format="%(asctime)s %(levelname)s %(name)s: %(message)s",
    )

    try:
        if args.command == "server" and args.server_command == "start":
            auth_hash = hash_passphrase(args.passphrase or "")
            asyncio.run(
                _run_server_start(
                    host=args.host,
                    port=args.port,
                    fixed_session=args.session,
                    fixed_auth_hash=auth_hash if args.session else None,
                    discovery_port=args.discovery_port,
                    enable_discovery=not args.no_discovery,
                )
            )
            return

        if args.command == "server" and args.server_command == "stop":
            _request_server_stop(wait_seconds=args.wait)
            return

        if args.command == "join":
            asyncio.run(
                _run_join(
                    server_host=args.host,
                    server_port=args.port,
                    session=args.session,
                    client_name=args.name,
                    passphrase=args.passphrase,
                    send_text=args.send,
                    sync_clipboard=not args.no_clipboard,
                    poll_interval=args.poll_interval,
                    heartbeat_interval=args.heartbeat_interval,
                    reconnect=not args.no_reconnect,
                    reconnect_delay=args.reconnect_delay,
                )
            )
            return

        if args.command == "host":
            asyncio.run(
                _run_host(
                    host=args.host,
                    port=args.port,
                    session=args.session,
                    passphrase=args.passphrase,
                    client_name=args.name,
                    send_text=args.send,
                    discovery_port=args.discovery_port,
                    no_discovery=args.no_discovery,
                    sync_clipboard=not args.no_clipboard,
                    poll_interval=args.poll_interval,
                    heartbeat_interval=args.heartbeat_interval,
                    reconnect=not args.no_reconnect,
                    reconnect_delay=args.reconnect_delay,
                )
            )
            return

        if args.command == "discover":
            sessions = asyncio.run(
                discover_sessions(
                    broadcast_host=args.broadcast,
                    discovery_port=args.port,
                    timeout_seconds=args.timeout,
                )
            )
            if not sessions:
                print("No sessions found.")
                return
            for item in sessions:
                locked = " [passphrase]" if item.requires_passphrase else ""
                print(f"{item.session} ({item.host}:{item.port}){locked}")
            return

        if args.command == "status":
            _print_status()
            return

        if args.command == "leave":
            _request_leave(wait_seconds=args.wait)
            return
    except KeyboardInterrupt:
        raise SystemExit(130)

    raise SystemExit("Unsupported command.")


def _build_parser() -> argparse.ArgumentParser:
    parser = argparse.ArgumentParser(prog="cliproom", description="ClipRoom CLI")
    parser.add_argument("--version", action="version", version=f"%(prog)s {__version__}")

    subparsers = parser.add_subparsers(dest="command")

    server_parser = subparsers.add_parser("server", help="Server control")
    server_sub = server_parser.add_subparsers(dest="server_command", required=True)

    start = server_sub.add_parser("start", help="Start CRP server")
    start.add_argument("--host", default="0.0.0.0", help="TCP bind host")
    start.add_argument("--port", type=int, default=DEFAULT_TCP_PORT, help="TCP port")
    start.add_argument(
        "--discovery-port",
        type=int,
        default=DEFAULT_DISCOVERY_PORT,
        help="UDP discovery port",
    )
    start.add_argument(
        "--no-discovery",
        action="store_true",
        help="Disable UDP discovery",
    )
    start.add_argument(
        "--session",
        help="Fixed session to allow. If omitted, server accepts multiple sessions.",
    )
    start.add_argument(
        "-p",
        "--passphrase",
        default="",
        help="Fixed session passphrase (only applies with --session).",
    )
    start.add_argument("-v", "--verbose", action="store_true", help="Verbose logs")

    stop = server_sub.add_parser("stop", help="Request local server stop")
    stop.add_argument(
        "--wait",
        type=float,
        default=5.0,
        help="Wait up to N seconds for server shutdown",
    )
    stop.add_argument("-v", "--verbose", action="store_true", help="Verbose logs")

    join = subparsers.add_parser("join", help="Join an existing session")
    join.add_argument("session", help="Session name")
    join.add_argument("--host", default="127.0.0.1", help="Server host")
    join.add_argument("--port", type=int, default=DEFAULT_TCP_PORT, help="TCP port")
    join.add_argument("--name", default="cliproom-client", help="Client name")
    join.add_argument("-p", "--passphrase", default="", help="Session passphrase")
    join.add_argument("--send", help="Send one initial text and then listen")
    join.add_argument(
        "--no-clipboard",
        action="store_true",
        help="Disable local clipboard monitoring",
    )
    join.add_argument(
        "--poll-interval",
        type=float,
        default=0.4,
        help="Clipboard polling interval in seconds",
    )
    join.add_argument(
        "--heartbeat-interval",
        type=float,
        default=10.0,
        help="Heartbeat interval in seconds",
    )
    join.add_argument(
        "--no-reconnect",
        action="store_true",
        help="Disable automatic reconnect",
    )
    join.add_argument(
        "--reconnect-delay",
        type=float,
        default=2.0,
        help="Reconnect delay in seconds",
    )
    join.add_argument("-v", "--verbose", action="store_true", help="Verbose logs")

    host = subparsers.add_parser("host", help="Start local server and join")
    host.add_argument("session", help="Session name")
    host.add_argument("--host", default="0.0.0.0", help="Server bind host")
    host.add_argument("--port", type=int, default=DEFAULT_TCP_PORT, help="TCP port")
    host.add_argument(
        "--discovery-port",
        type=int,
        default=DEFAULT_DISCOVERY_PORT,
        help="UDP discovery port",
    )
    host.add_argument(
        "--no-discovery",
        action="store_true",
        help="Disable UDP discovery",
    )
    host.add_argument("--name", default="cliproom-host", help="Local client name")
    host.add_argument("-p", "--passphrase", default="", help="Session passphrase")
    host.add_argument("--send", help="Send one initial text and then listen")
    host.add_argument(
        "--no-clipboard",
        action="store_true",
        help="Disable local clipboard monitoring",
    )
    host.add_argument(
        "--poll-interval",
        type=float,
        default=0.4,
        help="Clipboard polling interval in seconds",
    )
    host.add_argument(
        "--heartbeat-interval",
        type=float,
        default=10.0,
        help="Heartbeat interval in seconds",
    )
    host.add_argument(
        "--no-reconnect",
        action="store_true",
        help="Disable automatic reconnect",
    )
    host.add_argument(
        "--reconnect-delay",
        type=float,
        default=2.0,
        help="Reconnect delay in seconds",
    )
    host.add_argument("-v", "--verbose", action="store_true", help="Verbose logs")

    discover = subparsers.add_parser("discover", help="Discover LAN sessions")
    discover.add_argument(
        "--broadcast",
        default="255.255.255.255",
        help="Broadcast address",
    )
    discover.add_argument(
        "--port",
        type=int,
        default=DEFAULT_DISCOVERY_PORT,
        help="UDP discovery port",
    )
    discover.add_argument(
        "--timeout",
        type=float,
        default=2.0,
        help="Response timeout in seconds",
    )
    discover.add_argument("-v", "--verbose", action="store_true", help="Verbose logs")

    subparsers.add_parser("status", help="Show local client status")

    leave = subparsers.add_parser("leave", help="Request local client to disconnect")
    leave.add_argument(
        "--wait",
        type=float,
        default=5.0,
        help="Wait up to N seconds for client shutdown",
    )

    return parser


async def _run_join(
    *,
    server_host: str,
    server_port: int,
    session: str,
    client_name: str,
    passphrase: str,
    send_text: str | None,
    sync_clipboard: bool,
    poll_interval: float,
    heartbeat_interval: float,
    reconnect: bool,
    reconnect_delay: float,
) -> None:
    cli_state = LocalCLIState()
    cli_state.clear_leave_requested()
    sent_initial = False
    while True:
        if cli_state.is_leave_requested():
            break
        client = ClipRoomClient(
            server_host=server_host,
            server_port=server_port,
            session=session,
            client_name=client_name,
            passphrase=passphrase,
        )
        try:
            cli_state.write_state(
                _build_state_payload(
                    mode="join",
                    status="connecting",
                    session=session,
                    server_host=server_host,
                    server_port=server_port,
                    client_name=client_name,
                    connected=False,
                    last_error=None,
                )
            )
            await client.connect()
            cli_state.write_state(
                _build_state_payload(
                    mode="join",
                    status="connected",
                    session=session,
                    server_host=server_host,
                    server_port=server_port,
                    client_name=client_name,
                    connected=True,
                    last_error=None,
                )
            )
            if send_text and not sent_initial:
                await client.send_clip_text(send_text)
                sent_initial = True
            run_task = asyncio.create_task(
                client.run_forever(
                    sync_clipboard=sync_clipboard,
                    poll_interval=poll_interval,
                    heartbeat_interval=heartbeat_interval,
                )
            )
            leave_task = asyncio.create_task(_wait_for_leave_request(cli_state))
            done, pending = await asyncio.wait(
                [run_task, leave_task],
                return_when=asyncio.FIRST_COMPLETED,
            )
            for task in pending:
                task.cancel()
            for task in pending:
                with suppress(asyncio.CancelledError):
                    await task
            if leave_task in done:
                logging.getLogger("cliproom.cli").info("Leave requested, disconnecting...")
                await client.close()
                with suppress(asyncio.CancelledError, Exception):
                    await run_task
                break
            exc = run_task.exception()
            if exc is not None:
                raise exc
            if not reconnect:
                break
            cli_state.write_state(
                _build_state_payload(
                    mode="join",
                    status="disconnected",
                    session=session,
                    server_host=server_host,
                    server_port=server_port,
                    client_name=client_name,
                    connected=False,
                    last_error=None,
                )
            )
            logging.getLogger("cliproom.cli").warning(
                "Disconnected from server. Reconnecting in %.1fs...",
                reconnect_delay,
            )
            await asyncio.sleep(reconnect_delay)
        except KeyboardInterrupt:
            raise
        except Exception as exc:
            cli_state.write_state(
                _build_state_payload(
                    mode="join",
                    status="error",
                    session=session,
                    server_host=server_host,
                    server_port=server_port,
                    client_name=client_name,
                    connected=False,
                    last_error=str(exc),
                )
            )
            if not reconnect:
                raise
            logging.getLogger("cliproom.cli").warning(
                "Connection error: %s. Reconnecting in %.1fs...",
                exc,
                reconnect_delay,
            )
            await asyncio.sleep(reconnect_delay)
        finally:
            await client.close()
    cli_state.write_state(
        _build_state_payload(
            mode="join",
            status="stopped",
            session=session,
            server_host=server_host,
            server_port=server_port,
            client_name=client_name,
            connected=False,
            last_error=None,
        )
    )
    cli_state.clear_leave_requested()


async def _run_host(
    *,
    host: str,
    port: int,
    session: str,
    passphrase: str,
    client_name: str,
    send_text: str | None,
    discovery_port: int,
    no_discovery: bool,
    sync_clipboard: bool,
    poll_interval: float,
    heartbeat_interval: float,
    reconnect: bool,
    reconnect_delay: float,
) -> None:
    auth_hash = hash_passphrase(passphrase)
    server = ClipRoomServer(
        host=host,
        port=port,
        fixed_session=session,
        fixed_auth_hash=auth_hash,
        discovery_port=discovery_port,
        enable_discovery=not no_discovery,
    )
    await server.start()

    connect_host = "127.0.0.1" if host in {"0.0.0.0", "::"} else host
    server_task = asyncio.create_task(server.serve_forever())
    try:
        state = LocalCLIState()
        state.write_state(
            _build_state_payload(
                mode="host",
                status="hosting",
                session=session,
                server_host=connect_host,
                server_port=port,
                client_name=client_name,
                connected=False,
                last_error=None,
            )
        )
        await _run_join(
            server_host=connect_host,
            server_port=port,
            session=session,
            client_name=client_name,
            passphrase=passphrase,
            send_text=send_text,
            sync_clipboard=sync_clipboard,
            poll_interval=poll_interval,
            heartbeat_interval=heartbeat_interval,
            reconnect=reconnect,
            reconnect_delay=reconnect_delay,
        )
    finally:
        server_task.cancel()
        with suppress(asyncio.CancelledError):
            await server_task
        await server.stop()
        state = LocalCLIState()
        state.write_state(
            _build_state_payload(
                mode="host",
                status="stopped",
                session=session,
                server_host=connect_host,
                server_port=port,
                client_name=client_name,
                connected=False,
                last_error=None,
            )
        )


async def _run_server_start(
    *,
    host: str,
    port: int,
    fixed_session: str | None,
    fixed_auth_hash: str | None,
    discovery_port: int,
    enable_discovery: bool,
) -> None:
    state = LocalCLIState()
    state.clear_server_stop_requested()
    server = ClipRoomServer(
        host=host,
        port=port,
        fixed_session=fixed_session,
        fixed_auth_hash=fixed_auth_hash,
        discovery_port=discovery_port,
        enable_discovery=enable_discovery,
    )
    await server.start()
    state.write_server_state(
        _build_server_state_payload(
            status="running",
            host=host,
            port=port,
            discovery_port=discovery_port,
            fixed_session=fixed_session,
            last_error=None,
        )
    )

    serve_task = asyncio.create_task(server.serve_forever(), name="server-serve")
    stop_task = asyncio.create_task(
        _wait_for_server_stop_request(state),
        name="server-stop-request",
    )
    try:
        done, pending = await asyncio.wait(
            [serve_task, stop_task],
            return_when=asyncio.FIRST_COMPLETED,
        )
        if stop_task in done:
            logging.getLogger("cliproom.cli").info("Server stop requested.")
            await server.stop()
            serve_task.cancel()
            with suppress(asyncio.CancelledError):
                await serve_task
        else:
            exc = serve_task.exception()
            if exc is not None:
                raise exc
    except Exception as exc:
        state.write_server_state(
            _build_server_state_payload(
                status="error",
                host=host,
                port=port,
                discovery_port=discovery_port,
                fixed_session=fixed_session,
                last_error=str(exc),
            )
        )
        raise
    finally:
        stop_task.cancel()
        with suppress(asyncio.CancelledError):
            await stop_task
        await server.stop()
        state.write_server_state(
            _build_server_state_payload(
                status="stopped",
                host=host,
                port=port,
                discovery_port=discovery_port,
                fixed_session=fixed_session,
                last_error=None,
            )
        )
        state.clear_server_stop_requested()


def _utc_now() -> str:
    return datetime.now(timezone.utc).isoformat()


def _build_state_payload(
    *,
    mode: str,
    status: str,
    session: str,
    server_host: str,
    server_port: int,
    client_name: str,
    connected: bool,
    last_error: str | None,
) -> dict[str, object]:
    return {
        "active": status not in {"stopped"},
        "mode": mode,
        "status": status,
        "session": session,
        "server_host": server_host,
        "server_port": server_port,
        "client_name": client_name,
        "connected": connected,
        "pid": os.getpid(),
        "last_error": last_error,
        "updated_at": _utc_now(),
    }


async def _wait_for_leave_request(state: LocalCLIState, interval: float = 0.5) -> None:
    while True:
        if state.is_leave_requested():
            return
        await asyncio.sleep(interval)


async def _wait_for_server_stop_request(state: LocalCLIState, interval: float = 0.5) -> None:
    while True:
        if state.is_server_stop_requested():
            return
        await asyncio.sleep(interval)


def _print_status() -> None:
    state = LocalCLIState()
    client = state.read_state()
    server = state.read_server_state()

    if client is None and server is None:
        print("No local runtime state found.")
        return

    print("[client]")
    if client is None:
        print("state: missing")
    else:
        _print_state_block(
            {
                "active": client.get("active", False),
                "mode": client.get("mode", "unknown"),
                "status": client.get("status", "unknown"),
                "session": client.get("session", "-"),
                "endpoint": f"{client.get('server_host', '-')}:{client.get('server_port', '-')}",
                "name": client.get("client_name", "-"),
                "pid": client.get("pid", 0),
                "updated_at": client.get("updated_at", "-"),
                "last_error": client.get("last_error"),
            }
        )

    print("[server]")
    if server is None:
        print("state: missing")
    else:
        _print_state_block(
            {
                "active": server.get("active", False),
                "mode": "server",
                "status": server.get("status", "unknown"),
                "session": server.get("fixed_session") or "-",
                "endpoint": f"{server.get('host', '-')}:{server.get('port', '-')}",
                "name": "cliproom-server",
                "pid": server.get("pid", 0),
                "updated_at": server.get("updated_at", "-"),
                "last_error": server.get("last_error"),
            }
        )


def _print_state_block(payload: dict[str, object]) -> None:
    pid = int(payload.get("pid", 0) or 0)
    running = is_pid_running(pid)
    print(f"active: {payload.get('active', False)}")
    print(f"mode: {payload.get('mode', 'unknown')}")
    print(f"status: {payload.get('status', 'unknown')}")
    print(f"session: {payload.get('session', '-')}")
    print(f"endpoint: {payload.get('endpoint', '-')}")
    print(f"name: {payload.get('name', '-')}")
    print(f"pid: {pid} ({'running' if running else 'not running'})")
    print(f"updated_at: {payload.get('updated_at', '-')}")
    last_error = payload.get("last_error")
    if last_error:
        print(f"last_error: {last_error}")


def _request_leave(*, wait_seconds: float) -> None:
    state = LocalCLIState()
    payload = state.read_state()
    if payload is None:
        print("No local client state found.")
        return
    if not payload.get("active", False):
        print("Local client is not active.")
        return

    state.set_leave_requested()
    print("Leave requested.")
    if wait_seconds <= 0:
        return

    loop = asyncio.new_event_loop()
    try:
        loop.run_until_complete(_wait_for_client_stop(timeout=wait_seconds))
    finally:
        loop.close()


def _request_server_stop(*, wait_seconds: float) -> None:
    state = LocalCLIState()
    payload = state.read_server_state()
    if payload is None:
        print("No local server state found.")
        return
    if payload.get("status") != "running":
        print("Local server is not running.")
        return

    state.set_server_stop_requested()
    print("Server stop requested.")
    if wait_seconds <= 0:
        return

    loop = asyncio.new_event_loop()
    try:
        loop.run_until_complete(_wait_for_server_shutdown(timeout=wait_seconds))
    finally:
        loop.close()


async def _wait_for_client_stop(*, timeout: float) -> None:
    state = LocalCLIState()
    deadline = asyncio.get_running_loop().time() + timeout
    while asyncio.get_running_loop().time() < deadline:
        payload = state.read_state() or {}
        pid = int(payload.get("pid", 0) or 0)
        active = bool(payload.get("active", False))
        if (not active) or (pid and not is_pid_running(pid)):
            print("Client stopped.")
            return
        await asyncio.sleep(0.2)
    print("Timed out waiting for client shutdown.")


async def _wait_for_server_shutdown(*, timeout: float) -> None:
    state = LocalCLIState()
    deadline = asyncio.get_running_loop().time() + timeout
    while asyncio.get_running_loop().time() < deadline:
        payload = state.read_server_state() or {}
        pid = int(payload.get("pid", 0) or 0)
        status = payload.get("status")
        if status in {"stopped", "error"} or (pid and not is_pid_running(pid)):
            print("Server stopped.")
            return
        await asyncio.sleep(0.2)
    print("Timed out waiting for server shutdown.")


def _build_server_state_payload(
    *,
    status: str,
    host: str,
    port: int,
    discovery_port: int,
    fixed_session: str | None,
    last_error: str | None,
) -> dict[str, object]:
    return {
        "active": status == "running",
        "status": status,
        "host": host,
        "port": port,
        "discovery_port": discovery_port,
        "fixed_session": fixed_session,
        "pid": os.getpid(),
        "last_error": last_error,
        "updated_at": _utc_now(),
    }


if __name__ == "__main__":
    main()



"""Shared typing aliases for protocol layer."""

from __future__ import annotations

from typing import TYPE_CHECKING, Any, Literal, TypeAlias

if TYPE_CHECKING:
    from .messages import (
        ClipMessage,
        DiscoverMessage,
        DiscoverReplyMessage,
        ErrorMessage,
        HeartbeatMessage,
        JoinAckMessage,
        JoinMessage,
        LeaveMessage,
    )

MessageType: TypeAlias = Literal[
    "join",
    "join_ack",
    "clip",
    "leave",
    "heartbeat",
    "error",
    "discover",
    "discover_reply",
]

JSONScalar: TypeAlias = str | int | float | bool | None
JSONValue: TypeAlias = JSONScalar | list["JSONValue"] | dict[str, "JSONValue"]
JSONObject: TypeAlias = dict[str, JSONValue]

if TYPE_CHECKING:
    CRPMessage: TypeAlias = (
        JoinMessage
        | JoinAckMessage
        | ClipMessage
        | LeaveMessage
        | HeartbeatMessage
        | ErrorMessage
        | DiscoverMessage
        | DiscoverReplyMessage
    )
else:
    CRPMessage: TypeAlias = Any



"""CRP contracts and helpers."""

from .codec import decode_message_line, encode_message, hash_passphrase, parse_message
from .constants import (
    CRP_VERSION,
    DEFAULT_DISCOVERY_PORT,
    DEFAULT_DISCOVERY_TIMEOUT_SECONDS,
    DEFAULT_TCP_PORT,
)
from .errors import ProtocolValidationError
from .messages import (
    ClipboardFormat,
    ClipMessage,
    DiscoverMessage,
    DiscoverReplyMessage,
    ErrorMessage,
    HeartbeatMessage,
    JoinAckMessage,
    JoinMessage,
    LeaveMessage,
)
from .types import CRPMessage, JSONObject, JSONValue, MessageType

__all__ = [
    "CRP_VERSION",
    "DEFAULT_TCP_PORT",
    "DEFAULT_DISCOVERY_PORT",
    "DEFAULT_DISCOVERY_TIMEOUT_SECONDS",
    "ProtocolValidationError",
    "ClipboardFormat",
    "JoinMessage",
    "JoinAckMessage",
    "ClipMessage",
    "LeaveMessage",
    "HeartbeatMessage",
    "ErrorMessage",
    "DiscoverMessage",
    "DiscoverReplyMessage",
    "CRPMessage",
    "MessageType",
    "JSONValue",
    "JSONObject",
    "parse_message",
    "encode_message",
    "decode_message_line",
    "hash_passphrase",
]



"""Async client runtime for CRP sessions."""

from __future__ import annotations

import asyncio
import base64
import logging
import uuid
from contextlib import suppress

from cliproom.clipboard import (
    ClipboardAdapter,
    ClipboardUnavailableError,
    create_default_adapter,
)
from cliproom.protocol import (
    ClipMessage,
    ClipboardFormat,
    ErrorMessage,
    HeartbeatMessage,
    JoinAckMessage,
    JoinMessage,
    LeaveMessage,
    decode_message_line,
    encode_message,
    hash_passphrase,
)


LOG = logging.getLogger("cliproom.client")


class ClipRoomClient:
    """Minimal CRP client for joining a session and exchanging clip events."""

    def __init__(
        self,
        *,
        server_host: str,
        server_port: int,
        session: str,
        client_name: str,
        passphrase: str = "",
    ) -> None:
        self.server_host = server_host
        self.server_port = server_port
        self.session = session
        self.client_name = client_name
        self.auth_hash = hash_passphrase(passphrase)
        self._reader: asyncio.StreamReader | None = None
        self._writer: asyncio.StreamWriter | None = None
        self._connected = False
        self._clipboard_adapter: ClipboardAdapter | None = None
        self._last_local_signature: tuple[tuple[str, str], ...] | None = None
        self._last_remote_signature: tuple[tuple[str, str], ...] | None = None
        self._last_polled_signature: tuple[tuple[str, str], ...] | None = None

    async def connect(self) -> None:
        self._reader, self._writer = await asyncio.open_connection(
            self.server_host,
            self.server_port,
        )
        join = JoinMessage(
            session=self.session,
            auth=self.auth_hash,
            client_name=self.client_name,
        )
        await self._send(encode_message(join))

        line = await self._reader.readline()
        if not line:
            raise RuntimeError("Server closed connection before join_ack")
        message = decode_message_line(line)
        if isinstance(message, JoinAckMessage):
            self._connected = True
            LOG.info("Joined session '%s' as '%s'", self.session, self.client_name)
            return
        if isinstance(message, ErrorMessage):
            raise RuntimeError(f"Server error: {message.message}")
        raise RuntimeError(f"Unexpected first message: {message.type}")

    async def send_clip_text(self, text: str) -> None:
        encoded = base64.b64encode(text.encode("utf-8")).decode("ascii")
        await self.send_clip_formats(
            [
                ClipboardFormat(
                    mime="text/plain",
                    encoding="utf-8",
                    data=encoded,
                )
            ]
        )

    async def send_clip_formats(self, formats: list[ClipboardFormat]) -> None:
        if not self._connected:
            raise RuntimeError("Client is not connected")
        clip = ClipMessage(
            session=self.session,
            message_id=str(uuid.uuid4()),
            formats=formats,
        )
        await self._send(encode_message(clip))
        self._last_local_signature = _formats_signature(formats)

    async def receive_forever(self) -> None:
        if self._reader is None:
            raise RuntimeError("Client is not connected")
        while True:
            line = await self._reader.readline()
            if not line:
                LOG.info("Server closed connection")
                return
            message = decode_message_line(line)
            if isinstance(message, ClipMessage):
                LOG.info(
                    "Received clip message_id=%s formats=%d",
                    message.message_id,
                    len(message.formats),
                )
                self._last_remote_signature = _formats_signature(message.formats)
                if self._clipboard_adapter is not None:
                    try:
                        await asyncio.to_thread(
                            self._clipboard_adapter.set_formats,
                            message.formats,
                        )
                    except ClipboardUnavailableError as exc:
                        LOG.warning("Clipboard write failed: %s", exc)
                else:
                    self._print_fallback(message.formats)
            elif isinstance(message, ErrorMessage):
                raise RuntimeError(f"Server error: {message.message}")

    async def run_forever(
        self,
        *,
        sync_clipboard: bool = True,
        poll_interval: float = 0.4,
        heartbeat_interval: float = 10.0,
    ) -> None:
        if sync_clipboard:
            self._init_clipboard_adapter()

        tasks = [asyncio.create_task(self.receive_forever(), name="receive")]
        tasks.append(
            asyncio.create_task(
                self._send_heartbeat_forever(interval=heartbeat_interval),
                name="heartbeat",
            )
        )
        if sync_clipboard and self._clipboard_adapter is not None:
            tasks.append(
                asyncio.create_task(
                    self._monitor_local_clipboard_forever(poll_interval=poll_interval),
                    name="clipboard-monitor",
                )
            )

        done, pending = await asyncio.wait(tasks, return_when=asyncio.FIRST_COMPLETED)
        for task in pending:
            task.cancel()
        for task in pending:
            with suppress(asyncio.CancelledError):
                await task
        for task in done:
            exc = task.exception()
            if exc is not None:
                raise exc

    async def _send_heartbeat_forever(self, *, interval: float) -> None:
        while True:
            await asyncio.sleep(interval)
            if not self._connected:
                return
            with suppress(Exception):
                await self._send(encode_message(HeartbeatMessage()))

    async def _monitor_local_clipboard_forever(self, *, poll_interval: float) -> None:
        if self._clipboard_adapter is None:
            return
        while True:
            try:
                formats = await asyncio.to_thread(self._clipboard_adapter.get_formats)
            except ClipboardUnavailableError as exc:
                LOG.warning("Clipboard read failed: %s", exc)
                await asyncio.sleep(poll_interval)
                continue
            if not formats:
                await asyncio.sleep(poll_interval)
                continue
            signature = _formats_signature(formats)
            if signature != self._last_polled_signature:
                self._last_polled_signature = signature
                if signature != self._last_remote_signature:
                    await self.send_clip_formats(formats)
            await asyncio.sleep(poll_interval)

    def _init_clipboard_adapter(self) -> None:
        if self._clipboard_adapter is not None:
            return
        try:
            self._clipboard_adapter = create_default_adapter()
            LOG.info("Clipboard sync enabled")
        except ClipboardUnavailableError as exc:
            LOG.warning("Clipboard sync disabled: %s", exc)
            self._clipboard_adapter = None

    async def close(self) -> None:
        if self._writer is None:
            return
        with suppress(Exception):
            await self._send(encode_message(LeaveMessage()))
        self._writer.close()
        with suppress(Exception):
            await self._writer.wait_closed()
        self._writer = None
        self._reader = None
        self._connected = False

    async def _send(self, payload: bytes) -> None:
        if self._writer is None:
            raise RuntimeError("Writer is not available")
        self._writer.write(payload)
        await self._writer.drain()

    def _print_fallback(self, formats: list[ClipboardFormat]) -> None:
        for item in formats:
            if item.mime in {"text/plain", "text/html", "application/x-file-list"}:
                with suppress(Exception):
                    print(base64.b64decode(item.data).decode("utf-8"))
                    return
        if formats:
            print(f"<received {formats[0].mime}>")


def _formats_signature(formats: list[ClipboardFormat]) -> tuple[tuple[str, str], ...]:
    return tuple((item.mime, item.data) for item in formats)
